unit mcmain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, math, stypes;

type
  GrowthParValue = array[1..6] of double;
  TFmMCmain = class(TForm)
    MmInstruct: TMemo;
    LblInitialParFile: TLabel;
    LblBatchFile: TLabel;
    EdInitialParFile: TEdit;
    EdBatchFile: TEdit;
    LblMeasFile: TLabel;
    EdMeasFile: TEdit;
    EdInitialDriverFile: TEdit;
    LblInitialDriverFile: TLabel;
    DlgOpenList: TOpenDialog;
    BtnRunMC: TButton;
    BtnCloseMC: TButton;
    DlgSaveSensOutput: TSaveDialog;
    LblParFracFile: TLabel;
    EdParFracFile: TEdit;
    procedure LblInitialParFileClick(Sender: TObject);
    procedure LblBatchFileClick(Sender: TObject);
    procedure LblMeasFileClick(Sender: TObject);
    procedure ChooseParFracFile(Sender: TObject);
    procedure LblParFracFileClick(Sender: TObject);
    procedure CalcNewValues(var state:statearray; var param:paramarray;
              step:double; var NewDirection:Boolean);
    procedure FormCreate(Sender: TObject);
    procedure BtnCloseMCClick(Sender: TObject);
    procedure LblInitialDriverFileClick(Sender: TObject);
    procedure SetupRun(currentstat: statearray; currentpar: paramarray;
              step:double; var NewDirection: Boolean);
    procedure BtnRunMCClick(Sender: TObject);
    procedure UpdateForm(Sender:TObject);
    FUNCTION  ran0(VAR idum: integer): double;
    procedure FormShow(Sender: TObject);
    function  CheckGrowthParameters(var parameters:paramarray):Boolean;
  private
    { Private declarations }
    parfracfilename, initialparfilename, initialdriverfilename: string;
    initialpar, parfrac, simpar, minpar, maxpar, dpar: paramarray;
    initialstat, statfrac, simstat, minstat, maxstat, dstat: statearray;
    initialproc: processarray;
    initialresid, gly: double;
    glv: array[1..97] of double;
    GrowthName: array[1..6] of string; //(aGC, bGC, kGC, aGN, bGN, kGN);
    TissueName: array[1..3] of string;// (F, W, R);
  public
    { Public declarations }
  end;

var
  FmMCmain: TFmMCmain;

implementation

uses sensmain, batchmain, frontend, calculate, fileio, equations, ReloadDlg;

{$R *.DFM}

procedure TFmMCmain.LblInitialParFileClick(Sender: TObject);
begin
 MainForm.ChooseParamFile(Sender);
 initialparfilename := paramfilename;
 UpdateForm(Sender);
end;

procedure TFmMCmain.LblBatchFileClick(Sender: TObject);
begin
 FmSensMain.ChooseBatchFile(Sender);
 UpdateForm(Sender);
end;

procedure TFmMCmain.LblMeasFileClick(Sender: TObject);
begin
 FmSensMain.ChooseMeasFile(Sender);
 UpdateForm(Sender);
end;

procedure TFmMCmain.ChooseParFracFile(Sender: TObject);
var
 temp:double;
begin
 // If the user typed directly in the box
 if Sender is TEdit then parfracfilename := EdParFracFile.text
 else
  begin  // Show the open file dialog
   MainForm.DlgOpenParam.Options := MainForm.DlgOpenParam.Options +
                                      [ofNoChangeDir];
    // First set the dialog box default filename to the current paramfile
   MainForm.DlgOpenParam.filename := parfracfilename;
    // If the user chooses OK in the dialog then set the new paramfilename
   if MainForm.DlgOpenParam.execute then
     begin
      parfracfilename := MainForm.DlgOpenParam.filename;
     end;
  end;

 if (parfracfilename = '') or (not FileExists(parfracfilename)) then
  begin
    MessageDlg('Invalid Parameter File: ' + parfracfilename + '. Please reenter.',
                  mtError, [mbOK], 0);
    parfracfilename := '';  // Set the filename to no value
  end
 else
  begin
   if (not ReadParamFile(parfracfilename,ModelDef.numparam,parfrac,ModelDef.numstate,
                         statfrac, temp)) then parfracfilename := '';

  end;

 EdParFracfile.text := parfracfilename;
 UpdateForm(Sender);  // Update the form
end;

procedure TFmMCmain.LblParFracFileClick(Sender: TObject);
begin
 ChooseParFracFile(Sender);
 UpdateForm(Sender);
end;

procedure TFmMCmain.CalcNewValues(var state:statearray;
              var param:paramarray; step:double; var NewDirection:Boolean);
var
 i, one:integer;
 zpar:paramarray;
 zstat:statearray;
 label tryagain;
begin
tryagain:
 one := 1;
 // Calculate new change in value
 if NewDirection then
  begin
   for i := 1 to ModelDef.numstate do
      dstat[i].value := ran0(one)*statfrac[i].value;
   for i := 1 to ModelDef.numparam do
      dpar[i].value :=  ran0(one)*parfrac[i].value;
  end;
 for i := 1 to ModelDef.numstate do
  begin
 // Transform current values to z space
    zstat[i].value := tan((pi/2)*(state[i].value - minstat[i].value)/
                      (maxstat[i].value - minstat[i].value));
 // New value in z space
    zstat[i].value := zstat[i].value*(1+step*dstat[i].value);
 // Convert back to real (but constrained) values
    state[i].value := minstat[i].value + (maxstat[i].value - minstat[i].value)*
         (2/pi)*arctan(zstat[i].value);
  end;
 for i := 1 to ModelDef.numparam do
  begin
 // Transform current values to z space
    zpar[i].value := tan((pi/2)*(param[i].value - minpar[i].value)/
                      (maxpar[i].value - minpar[i].value));
 // New value in z space
    zpar[i].value := zpar[i].value*(1+step*dpar[i].value);
 // Convert back to real (but constrained) values
    param[i].value := minpar[i].value + (maxpar[i].value - minpar[i].value)*
         (2/pi)*arctan(zpar[i].value);
  end;
 if not CheckGrowthParameters(param) then
  begin
   NewDirection := True;
   Goto tryagain;
  end;
end;

procedure TFmMCmain.FormCreate(Sender: TObject);
begin
 Randomize;
 GrowthName[1] := 'aGC';
 GrowthName[2] := 'bGC';
 GrowthName[3] := 'kGC';
 GrowthName[4] := 'aGN';
 GrowthName[5] := 'bGN';
 GrowthName[6] := 'kGN';
 TissueName[1] := 'F';
 TissueName[2] := 'W';
 TissueName[3] := 'R';
end;

procedure TFmMCmain.BtnCloseMCClick(Sender: TObject);
begin
 FmMCmain.Close;
end;

procedure TFmMCmain.LblInitialDriverFileClick(Sender: TObject);
begin
 MainForm.ChooseDriver(Sender);
 InitialDriverfilename := driverfilename;
 UpdateForm(Sender);
end;

procedure TFmMCmain.SetupRun(currentstat: statearray; currentpar: paramarray;
             step:double; var NewDirection: Boolean);
var
 Path : string;
 i: integer;
begin
// Calculate new values
 CalcNewValues(currentstat, currentpar, step, NewDirection);
// Calculate a steadystate
 try
// Need to replace driver with inital driver b/c the driver file will change during the batch job driverfilename := initialdriverfilename;
   driverfilename := initialdriverfilename;
   FmCalculate.FormShow(FmMCmain);
   for i := 1 to ModelDef.numstate do
    begin
      if statfrac[i].value <> 0 then
         FmCalculate.ChangeValue(vtstate, i, currentstat[i].value, false);
    end;
   for i := 1 to ModelDef.numparam do
    begin
      if parfrac[i].value <> 0 then
         FmCalculate.ChangeValue(vtparameter, i, currentpar[i].value, false);
    end;
   FmCalculate.ChangeValue(vtparameter, 77, currentpar[77].value, true);
   FmCalculate.BtnOKClick(FmMCmain); // Transfers new values into stat, par, and proc
   FmCalculate.Hide;
{ Save the new parameter set in file montecarlo.par. Remember that the batch
   must specify montecarlo.par as the parameter filename. }
   Path := ExtractFilePath(initialparfilename);
   paramfilename := Path + 'montecarlo.par';
   MainForm.MISaveParamClick(FmMCmain);
 except
   raise;
 end;
end;

procedure TFmMCmain.BtnRunMCClick(Sender: TObject);
label StartHere;
var
 simresid, minfrac, newresid, minstep, initialstep, step: double;
 savecount, i, minusone: integer;
 path: string;
 NewDirection: Boolean;
begin
 minusone := -1;
 BtnRunMC.enabled := False;
 initialpar := par;   // Just to get names and symbols
 initialstat := stat;   // Just to get names and symbols
 initialproc := proc;    // "
 ReadParamFile(initialparfilename, ModelDef.numparam, initialpar,
       ModelDef.numstate, initialstat, initialresid);
 simpar := initialpar;
 simstat := initialstat;
 simresid := initialresid;
// Load initial values into stat and par for steadystate calculation
 par := initialpar;
 stat := initialstat;
 savecount := 1;
 Path := ExtractFilePath(initialparfilename);
 minfrac := 999;
 for i := 1 to modeldef.numstate do
  begin
   statfrac[i].value := 0.9*statfrac[i].value;
   if (statfrac[i].value > 0) and (statfrac[i].value < minfrac)
             then minfrac := statfrac[i].value;
  end;
 for i := 1 to modeldef.numparam do
  begin
   parfrac[i].value := 0.9*parfrac[i].value;
   if (parfrac[i].value > 0) and (parfrac[i].value < minfrac)
             then minfrac := parfrac[i].value;
  end;
 initialstep := 0.9/minfrac;
 minstep := initialstep/1000;
StartHere:
 ran0(minusone);
 step := initialstep;
 NewDirection := True;
 while abs(step) > minstep do
  begin
    try
    // Get parameters, states and drivers for simulation.
     SetupRun(simstat, simpar, step, NewDirection);
    // Run model runs listed in the batch file
     FmBatchMain.BtnRunBatchClick(Sender);
    // Get the results of the test runs and calculate the residuals
     FmSensMain.GetMeasurements;
     FmSensMain.GetSimResults;
     FmSensMain.CalculateResiduals;
     newresid := FmSensMain.ResidCT + FmSensMain.ResidNT;   // residNT includes residNEP
     DlgReload.OKBtnClick(Sender);   // Reload state variables
     if newresid < simresid then
      begin
       simpar := par;
       simstat := stat;
       simresid := newresid;
       paramfilename := 'currentmc.par';
       WriteParamFile(paramfilename, ModelDef.numparam, simpar,
                              ModelDef.numstate, simstat, simresid);
       NewDirection := False;
      end
     else
      begin
       step := -step/2.0;
       NewDirection := False;
      end;
    except
     DlgReload.OKBtnClick(Sender);   // Reload state variables
     NewDirection := True;
     step := step*0.999;
    end;
  end;
 paramfilename := 'montecarlo' + inttostr(savecount) + '.par';
 WriteParamFile(paramfilename, ModelDef.numparam, simpar,
                              ModelDef.numstate, simstat, simresid);
 savecount := savecount + 1;
{ simpar := initialpar;
 simstat := initialstat;
 simresid := initialresid;  }
 goto StartHere;
 BtnRunMC.enabled := True;       // fix? - code never gets here
end;

procedure TFmMCmain.UpdateForm(Sender:TObject);
begin
 if (InitialParFilename <> '') and (InitialDriverFilename <> '') and
   (parfracfilename <> '') and (BatchFilename <> '')
   and (FmSensMain.MeasFilename <> '')then
   begin
     BtnRunMC.Enabled := True;
     FmMCMain.ActiveControl := BtnRunMC;
   end;
end;

FUNCTION TFmMCmain.ran0(VAR idum: integer): double;
(* RAN0 assumes that the system random number generator is called RANDOM,
with no arguments. Programs using RAN0 must declare the following
variables
VAR
   gly: real;
   glv: ARRAY [1..97] OF real;
in the main routine. *)
VAR
   dum: double;
   j: integer;
BEGIN
   IF (idum < 0) THEN BEGIN
      idum := 1;
      FOR j := 1 to 97 DO dum := random;
      FOR j := 1 to 97 DO glv[j] := random;
      gly := random;
   END;
   j := 1+trunc(97.0*gly);
   IF ((j > 97) OR (j < 1)) THEN BEGIN
      raise Exception.Create('Error in ran0'); {writeln('pause in routine RAN0'); readln} END;
   gly := glv[j];
   ran0 := gly;
   glv[j] := random;
END;

procedure TFmMCmain.FormShow(Sender: TObject);
var
 i:integer;
begin
 // For names and units
 minstat := stat;
 maxstat := stat;
 dstat := stat;
 minpar := par;
 maxpar := par;
 dpar := par;
 for i := 1 to ModelDef.numstate do
   begin
     if (i < 7) or (i > 12) then
      begin
       minstat[i].value := 0;
       maxstat[i].value := 1E10;
      end
     else if i > 9 then // 10 <= i <= 12     Labile N Pools
      begin
       minstat[i].value := 0;
       maxstat[i].value := 5;
      end
     else  // 7 <= i < 9      Labile C Pools
      begin
       minstat[i].value := 0;
       maxstat[i].value := 5;
      end;
   end;
 for i := 1 to ModelDef.numparam do
   begin
     minpar[i].value := 0;
     maxpar[i].value := 1e8;
   end;
// mf
 i := FmCalculate.GetArrayIndex(vtparameter,'mf');
 minpar[i].value := 70;
 maxpar[i].value := 90;
// sigmaF, sigmaW, sigmaR
 i := FmCalculate.GetArrayIndex(vtparameter,'sigmaF');
 minpar[i].value := 0.5;
 maxpar[i].value := 2;
 i := FmCalculate.GetArrayIndex(vtparameter,'sigmaW');
 minpar[i].value := 0.5;
 maxpar[i].value := 2;
 i := FmCalculate.GetArrayIndex(vtparameter,'sigmaR');
 minpar[i].value := 0.5;
 maxpar[i].value := 2;
// qu0
 i := FmCalculate.GetArrayIndex(vtparameter,'qu0');
 minpar[i].value := 1;
 maxpar[i].value := 100;
// aWR
 i := FmCalculate.GetArrayIndex(vtparameter,'aWR');
 minpar[i].value := 0.4;
 maxpar[i].value := 0.6;
// Trmax and Tropt
 i := FmCalculate.GetArrayIndex(vtparameter,'Trmax');
 minpar[i].value := 20;
 maxpar[i].value := 100;
 i := FmCalculate.GetArrayIndex(vtparameter,'Tropt');
 minpar[i].value := 1;
 maxpar[i].value := 100;
// Q10s
 i := FmCalculate.GetArrayIndex(vtparameter,'Q10s');
 minpar[i].value := 1;
 maxpar[i].value := 5;
// beta
 i := FmCalculate.GetArrayIndex(vtparameter,'beta');
 minpar[i].value := 0.1;
 maxpar[i].value := 0.39;
// Ase
 i := FmCalculate.GetArrayIndex(vtparameter,'Ase');
 minpar[i].value := 0.00001;
 maxpar[i].value := 2;
// Aei
 i := FmCalculate.GetArrayIndex(vtparameter,'Aei');
 minpar[i].value := 0.00001;
 maxpar[i].value := 3;
// Aie
 i := FmCalculate.GetArrayIndex(vtparameter,'Aie');
 minpar[i].value := 0.00001;
 maxpar[i].value := 2;
// Aih
 i := FmCalculate.GetArrayIndex(vtparameter,'Aih');
 minpar[i].value := 0.00001;
 maxpar[i].value := 2;
// qI
 i := FmCalculate.GetArrayIndex(vtparameter,'qI');
 minpar[i].value := 10;
 maxpar[i].value := 60;
// Wdmin
 i := FmCalculate.GetArrayIndex(vtparameter,'Wdmin');
 minpar[i].value := 10;
 maxpar[i].value := 1000;
// Wdopt
 i := FmCalculate.GetArrayIndex(vtparameter,'Wdopt');
 minpar[i].value := 50;
 maxpar[i].value := 1000;
// JWdmax
 i := FmCalculate.GetArrayIndex(vtparameter,'JWdmax');
 minpar[i].value := 0;
 maxpar[i].value := 1;
// phiD
 i := FmCalculate.GetArrayIndex(vtparameter,'phiD');
 minpar[i].value := 0;
 maxpar[i].value := 1;
end;

function TFmMCmain.CheckGrowthParameters(var parameters:paramarray):Boolean;
var
 i,j,k:integer;
 Gpar:GrowthParValue;
{ aGCF, bGCF, kGCF, aGNF, bGNF, kGNF, aGCW, bGCW, kGCW, aGNW, bGNW, kGNW,
 aGCR, bGCR, kGCR, aGNR, bGNW, kGNR: double;  }
begin
CheckGrowthParameters := True;
for j := 1 to 3 do
 begin
  for i := 1 to 6 do
   begin
     k := FmCalculate.GetArrayIndex(vtparameter,GrowthName[i]+TissueName[j]);
     Gpar[i] := parameters[k].value;
   end;
  if not(Gpar[4] > Gpar[1]) then {aGN > aGC}
    CheckGrowthParameters := False;
  if not(Gpar[5] < Gpar[2]) then {bGN < bGC}
    CheckGrowthParameters := False;
  if not(Gpar[6]*Gpar[1] < Gpar[3]*Gpar[4]) then {kGNaGC < kGCaGN}
    CheckGrowthParameters := False;
  if not(Gpar[3]*Gpar[5] < Gpar[6]*Gpar[2]) then {kGCbGN < kGNbGC}
    CheckGrowthParameters := False;
 end;
end;

{
// Foliage
 i := FmCalculate.GetArrayIndex(vtparameter,'aGCF');
 aGCF := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'bGCF');
 bGCF := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'kGCF');
 kGCF := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'aGNF');
 aGNF := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'bGNF');
 bGNF := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'aGNF');
 kGNF := parameters[i].value;
// Wood
 i := FmCalculate.GetArrayIndex(vtparameter,'aGCW');
 aGCW := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'bGCW');
 bGCW := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'kGCW');
 kGCW := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'aGNW');
 aGNW := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'bGNW');
 bGNW := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'aGNW');
 kGNW := parameters[i].value;
// Root
 i := FmCalculate.GetArrayIndex(vtparameter,'aGCR');
 aGCR := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'bGCR');
 bGCR := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'kGCR');
 kGCR := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'aGNR');
 aGNR := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'bGNR');
 bGNR := parameters[i].value;
 i := FmCalculate.GetArrayIndex(vtparameter,'aGNR');
 kGNR := parameters[i].value;

}

end.
